from pathlib import Path

import torch
import torchaudio
from torch.utils.data import DataLoader, Dataset
from torch.utils.data import random_split
from transformers import AutoFeatureExtractor

from helper_code import load_patient_data, get_age, get_sex, get_height, get_weight, get_pregnancy_status, get_murmur, \
    get_outcome

NUM_WORKERS = 6


class CirCorDigiScopePCG(Dataset):
    def __init__(self, root):
        self.root = root
        self._walker = list(Path(root).glob('*.wav'))
        self.murmur_classes = ['Present', 'Unknown', 'Absent']
        self.outcome_classes = ['Abnormal', 'Normal']

        self._murmur_class2idx = {'Present': 0, 'Unknown': 1, 'Absent': 2}
        self._outcome_class2idx = {'Abnormal': 0, 'Normal': 1}

    def __len__(self):
        return len(self._walker)

    def __getitem__(self, n):
        filepath = self._walker[n]
        waveform, sample_rate = torchaudio.load(filepath)
        length = waveform.shape[1]
        if length > sample_rate * 20:
            waveform = waveform[:, 10*sample_rate:20*sample_rate]
        waveform, sample_rate = torchaudio.functional.resample(waveform, sample_rate, 16000), 16000

        patient_filename = filepath.stem.split('_')[0] + '.txt'
        patient_data = load_patient_data(Path(self.root) / patient_filename)

        return {"waveform": waveform.squeeze().numpy(),
                "sample_rate": sample_rate,
                "murmur": self._murmur_class2idx[get_murmur(patient_data)],
                "outcome": self._outcome_class2idx[get_outcome(patient_data)],
                "age": get_age(patient_data),
                "sex": get_sex(patient_data),
                "height": get_height(patient_data),
                "weight": get_weight(patient_data),
                "pregnancy_status": get_pregnancy_status(patient_data),
                }


# class CollateFn:
#     def __init__(self):
#         self.feature_extractor = AutoFeatureExtractor.from_pretrained("facebook/wav2vec2-base-960h")
#
#     def __call__(self, batch):
#         return self.feature_extractor([d["waveform"] for d in batch], padding=True, return_tensors='pt', sampling_rate=16000), \
#                torch.LongTensor([d["murmur"] for d in batch])

class CollateFn:
    def __init__(self):
        self.feature_extractor = AutoFeatureExtractor.from_pretrained("facebook/wav2vec2-base-960h")

    def __call__(self, batch):
        return {
            "murmur": torch.LongTensor([d["murmur"] for d in batch]),
            "outcome": torch.LongTensor([d["outcome"] for d in batch]),
            **self.feature_extractor([d["waveform"] for d in batch], padding=True, return_tensors='pt', sampling_rate=16000)
        }


class PCGDataModule:
    def __init__(self, batch_size, root, val_split=True):
        self.batch_size = batch_size

        data = CirCorDigiScopePCG(root)
        if val_split:
            self.pcg_train, self.pcg_val = random_split(data, [3000, 163])
        else:
            self.pcg_train = data

    def train_dataloader(self):
        return DataLoader(self.pcg_train, batch_size=self.batch_size, num_workers=NUM_WORKERS, pin_memory=True,
                          drop_last=False, shuffle=True, collate_fn=CollateFn())

    def val_dataloader(self):
        return DataLoader(self.pcg_val, batch_size=self.batch_size, num_workers=NUM_WORKERS, pin_memory=True,
                          drop_last=False, collate_fn=CollateFn())

    def test_dataloader(self):
        pass
