import argparse
import os
from pathlib import Path

import datasets
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchaudio
import torchmetrics
from torch.utils.data import DataLoader, random_split
from transformers import Wav2Vec2Model

from data import CirCorDigiScopePCG, CollateFn
from helper_code import load_patient_data, get_murmur

NUM_WORKERS = 8


# def create_dataset(root):
#     waveforms = []
#     for filepath in Path(root).glob('*.wav'):
#         waveform, sample_rate = torchaudio.load(filepath)
#         waveform = torchaudio.functional.resample(waveform, sample_rate, 16000)
#         waveforms.append(waveform.squeeze().numpy())
#
#     feature_extractor = AutoFeatureExtractor.from_pretrained("facebook/wav2vec2-base-960h")
#     waveforms_tensor = feature_extractor(
#         waveforms, padding=True, return_tensors='pt', sampling_rate=16000)['input_values']
#
#     return TensorDataset(waveforms_tensor)


def create_dataset(root):
    data = {
        "path": [],
        "murmur": [],
    }
    features = datasets.Features({
        'path': datasets.Value(dtype='string'),
        'murmur': datasets.ClassLabel(num_classes=3, names=['Present', 'Unknown', 'Absent']),
    })

    def get_waveform(filepath):
        waveform, sample_rate = torchaudio.load(filepath)
        waveform = torchaudio.functional.resample(waveform, sample_rate, 16000)
        return waveform.squeeze().numpy()

    def append_waveform(examples):
        examples["waveform"] = [get_waveform(path) for path in examples["path"]]
        return examples

    for filepath in Path(root).glob('*.wav'):
        patient_filename = filepath.stem.split('_')[0] + '.txt'
        patient_data = load_patient_data(Path(root) / patient_filename)
        murmur = get_murmur(patient_data)

        data["path"].append(str(filepath))
        data["murmur"].append(murmur)

    return datasets.Dataset.from_dict(data, features=features).map(append_waveform, batched=True, batch_size=50). \
        remove_columns(['path'])


class PCGLitDataModule(pl.LightningDataModule):
    def __init__(self, batch_size, root, val_split=True):
        super().__init__()
        self.batch_size = batch_size
        self.root = root
        self.val_split = val_split

    def setup(self, stage=None):
        data = CirCorDigiScopePCG(self.root)
        # data = create_dataset(self.root)
        if self.val_split:
            self.pcg_train, self.pcg_val = random_split(data, [3000, 163])
            # data = data.train_test_split(test_size=163)
            # self.pcg_train, self.pcg_val = data['train'], data['test']
        else:
            self.pcg_train = data

    def train_dataloader(self):
        return DataLoader(self.pcg_train, batch_size=self.batch_size, num_workers=NUM_WORKERS, pin_memory=True,
                          drop_last=False, shuffle=True, collate_fn=CollateFn())

    def val_dataloader(self):
        return DataLoader(self.pcg_val, batch_size=self.batch_size, num_workers=NUM_WORKERS, pin_memory=True,
                          drop_last=False, collate_fn=CollateFn()) if self.val_split else None


class Wav2Vec2Classifier(nn.Module):
    def __init__(self):
        super().__init__()
        self.wav2vec = Wav2Vec2Model.from_pretrained("facebook/wav2vec2-base-960h")
        # for param in self.wav2vec.parameters():
        #    param.requires_grad = False
        self.wav2vec.feature_extractor._freeze_parameters()
        self.linear_head = nn.Linear(768, 3)

    def forward(self, x):
        hidden_state = self.wav2vec(x)[0]
        pooled_hidden_state = torch.mean(hidden_state, dim=1)
        logits = self.linear_head(pooled_hidden_state)
        return logits


class Wav2VecMurmurLitModel(pl.LightningModule):
    def __init__(self, opt):
        super().__init__()
        self.opt = opt
        self.model = Wav2Vec2Classifier()
        self.accuracy = torchmetrics.Accuracy()

    def forward(self, batch):
        logits = self.model(batch['input_values'])
        return logits

    def training_step(self, batch, batch_idx):
        logits = self.model(batch['input_values'])
        loss = F.cross_entropy(logits, batch['murmur'])
        return loss

    def validation_step(self, batch, batch_idx):
        logits = self.model(batch['input_values'])
        loss = F.cross_entropy(logits, batch['murmur'])
        self.accuracy(logits, batch['murmur'])
        self.log("val_loss", loss, prog_bar=True, sync_dist=True)
        self.log("val_acc", self.accuracy, prog_bar=True, sync_dist=True)

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.opt.learning_rate, weight_decay=self.opt.weight_decay,
                                amsgrad=True)


def main(opt):
    pl.seed_everything(opt.seed)
    root = os.path.join(
        os.environ['SLURM_TMPDIR'], 'datasets/the-circor-digiscope-phonocardiogram-dataset-1.0.3/training_data') \
        if 'SLURM_TMPDIR' in os.environ else 'data/training_data'
    dm = PCGLitDataModule(batch_size=opt.batch_size, root=root, val_split=True)
    model = Wav2VecMurmurLitModel(opt)

    trainer = pl.Trainer(
        devices=-1,
        accelerator="gpu",
        strategy="ddp",
        precision=16,
        max_epochs=20,
        default_root_dir=os.getcwd(),
    )
    trainer.fit(model, datamodule=dm)


def parse_option():
    parser = argparse.ArgumentParser('argument for training')

    parser.add_argument('--seed', type=int, default=0,
                        help='seed')
    parser.add_argument('--batch_size', type=int, default=2,
                        help='batch_size')

    # optimization
    parser.add_argument('--learning_rate', type=float, default=1e-4,
                        help='learning rate')
    parser.add_argument('--weight_decay', type=float, default=5e-4,
                        help='weight decay')
    parser.add_argument('--momentum', type=float, default=0,
                        help='momentum')

    opt = parser.parse_args()

    return opt


if __name__ == '__main__':
    opt = parse_option()
    print(opt)
    main(opt)
