# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the
# required functions, change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################
import argparse
from pathlib import Path

import numpy as np
import pytorch_lightning as pl
import torch
import torchaudio
from transformers import AutoFeatureExtractor

from mumur import PCGLitDataModule, Wav2VecMurmurLitModel
from outcome import Wav2VecOutcomeLitModel


################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.


def train_challenge_model(data_folder, model_folder, verbose):
    pl.seed_everything(0)
    opt = argparse.Namespace()
    opt.learning_rate = 1e-4
    opt.weight_decay = 5e-4

    dm = PCGLitDataModule(batch_size=1, root=data_folder, val_split=False)
    murmur_model = Wav2VecMurmurLitModel(opt)

    murmur_trainer = pl.Trainer(
        devices=1,
        accelerator="gpu",
        precision=16,
        max_epochs=18,
        enable_checkpointing=False,
        default_root_dir=model_folder,
    )
    murmur_trainer.fit(murmur_model, datamodule=dm)
    murmur_trainer.save_checkpoint(Path.cwd() / Path(model_folder) / 'murmur_model.ckpt')

    opt.weight_decay = 0.
    outcome_model = Wav2VecOutcomeLitModel(opt)

    outcome_trainer = pl.Trainer(
        devices=1,
        accelerator="gpu",
        precision=16,
        max_epochs=2,
        enable_checkpointing=False,
        default_root_dir=model_folder,
    )
    outcome_trainer.fit(outcome_model, datamodule=dm)
    outcome_trainer.save_checkpoint(Path.cwd() / Path(model_folder) / 'outcome_model.ckpt')

    if verbose >= 1:
        print('Done.')


# Load your trained model. This function is *required*. You should edit this function to add your code,
# but do *not* change the arguments of this function.
def load_challenge_model(model_folder, verbose):
    return {
        "murmur_model": Wav2VecMurmurLitModel.load_from_checkpoint(Path.cwd() / Path(model_folder) / 'murmur_model.ckpt', opt=1),
        "outcome_model": Wav2VecOutcomeLitModel.load_from_checkpoint(Path.cwd() / Path(model_folder) / 'outcome_model.ckpt', opt=1),
    }


# Run your trained model. This function is *required*. You should edit this function to add your code,
# but do *not* change the arguments of this function.
def run_challenge_model(model, data, recordings, verbose):
    # load models
    feature_extractor = AutoFeatureExtractor.from_pretrained("facebook/wav2vec2-base-960h")
    murmur_model = model['murmur_model']
    outcome_model = model['outcome_model']

    recordings = [torchaudio.functional.resample(torch.from_numpy(recording) / 32768., 4000, 16000).numpy() for
                  recording in recordings]

    murmur_logits = murmur_model(feature_extractor(recordings, padding=True, return_tensors='pt', sampling_rate=16000))
    murmur_probability = torch.nn.functional.softmax(murmur_logits, dim=1).mean(0).detach().numpy()
    murmur_labels = np.zeros(3, dtype=int)
    murmur_labels[np.argmax(murmur_probability)] = 1

    outcome_logits = outcome_model(feature_extractor(recordings, padding=True, return_tensors='pt', sampling_rate=16000))
    outcome_probability = torch.nn.functional.softmax(outcome_logits, dim=1).mean(0).detach().numpy()
    outcome_labels = np.zeros(2, dtype=int)
    outcome_labels[np.argmax(outcome_probability)] = 1

    # Concatenate classes, labels, and probabilities.
    murmur_classes = ['Present', 'Unknown', 'Absent']
    outcome_classes = ['Abnormal', 'Normal']
    classes = murmur_classes + outcome_classes
    labels = np.concatenate((murmur_labels, outcome_labels))
    probabilities = np.concatenate((murmur_probability, outcome_probability))

    return classes, labels, probabilities
